unit Helios_ROQ;
{
 * ------------------------------------------------------------------------
 * Title: RoQ Video File Delphi Conversion
 * Ver:   v1.0
 * By:    Stuart Carey (Stucuk)
 *
 * Convertion Start: 18/09/04
 * Comments:
 *
 *
 * ------------------------------------------------------------------------
 * Id Software's RoQ video file format decoder
 *
 * Dr. Tim Ferguson, 2001.
 * For more details on the algorithm:
 *         http://www.csse.monash.edu.au/~timf/videocodec.html
 *
 * This is a simple decoder for the Id Software RoQ video format.  In
 * this format, audio samples are DPCM coded and the video frames are
 * coded using motion blocks and vector quantisation.
 *
 * Note: All information on the RoQ file format has been obtained through
 *   pure reverse engineering.  This was achieved by giving known input
 *   audio and video frames to the roq.exe encoder and analysing the
 *   resulting output text and RoQ file.  No decompiling of the Quake III
 *   Arena game was required.
 *
 * You may freely use this source code.  I only ask that you reference its
 * source in your projects documentation:
 *       Tim Ferguson: http://www.csse.monash.edu.au/~timf/
 * ------------------------------------------------------------------------
}


interface

const
RoQ_INFO          = $1001;
RoQ_QUAD_CODEBOOK = $1002;
RoQ_QUAD_VQ       = $1011;
RoQ_SOUND_MONO    = $1020;
RoQ_SOUND_STEREO  = $1021;

RoQ_ID_MOT = $00;
RoQ_ID_FCC = $01;
RoQ_ID_SLD = $02;
RoQ_ID_CCC = $03;

//New
SEEK_CUR = 1;
SEEK_SET = 2;

type
Proq_cell = ^roq_cell;
roq_cell = record
 y0,y1,y2,y3,u,v : Byte;
end;

type
Proq_qcell = ^roq_qcell;
roq_qcell = record
idx : array [0..3] of integer;
end;

Type
PFile = ^File;

type
Proq_info = ^Troq_info;
Troq_info = record
  FP : PFile;
  Buf_Size : Integer;
  Buf : array of ^Byte;
  cells : array [0..255] of roq_cell;
  qcells : array [0..255] of roq_qcell;
  snd_sqr_arr : array [0..259] of Shortint;
  roq_start, aud_pos, vid_pos : Longint;
  Frame_offset : array of Longint;
  num_frames,num_audio_bytes : longword;
  width, height, frame_num, audio_channels : integer;
  y,u,v : array [0..1] of array of Byte;
  stream_length : longint;
  audio_buf_size, audio_size : integer;
  audio : array of Byte;
end;

Function roq_open(fname : pchar) : Proq_info;
Procedure roq_close(RI : Proq_info);
Function roq_read_audio(Ri : Proq_info) : integer;

implementation

uses Windows, Sysutils;

// Delphi's Equivalent to C Procedures/Functions

// New
Function FGetC(FP : PFile) : Byte;
begin
BlockRead(FP^,Result,Sizeof(Byte)); // Read Byte
end;

//New
Function FTell(FP : PFile) : Longint;
begin
Result := FilePos(FP^);
end;

//New
Function Feof(FP : PFile) : Boolean;
begin
Result := Eof(FP^);
end;

//New
Function Fseek(FP : PFile; Offset : Longint; whence : integer) : Boolean;
var
Pos : Longint;
begin

Pos := Offset;
if whence = SEEK_CUR then
Pos := Pos + FTell(FP);

Seek(FP^,Pos);
end;

// Start of roq_read.h
Function Get_Word(FP : PFile) : Word;
begin
Result := (FGetC(FP) and $FF);
Result := Result or (FGetC(FP) and $FF) shl 8;
end;

Function Get_Long(FP : PFile) : Longword;
begin
Result := (FGetC(FP) and $FF);
Result := Result or (FGetC(FP) and $FF) shl 8;
Result := Result or (FGetC(FP) and $FF) shl 16;
Result := Result or (FGetC(FP) and $FF) shl 24;
end;

// Comment: Max_Frame removed, isn't needed since setlength replaces the mem management of the dynamic array
Function roq_parse_file(FP : PFile; RI : Proq_info) : integer;
var
chunk_id,chunk_arg : Cardinal;
chunk_size, fpos : longint;
head1,head3 : Word;
head2 : LongWord;
begin
RI^.num_frames := 0;
RI^.num_audio_bytes := 0;
RI^.audio_channels := 0;
RI^.Frame_offset := nil;
RI^.Buf_Size := 0;

head1 := get_word(fp);
head2 := get_long(fp);
head3 := get_word(fp);

if(not (head1 = $1084) and not (head2 = $ffffffff) and not (head3 = $1e)) then
begin
MessageBox(0,'Not an RoQ file','Invalid Header',0);
Result := 1;
Exit;
end;

RI^.roq_start := FTell(fp);

while not Feof(FP) do
begin

fpos       := FTell(FP);
chunk_id   := get_word(fp);
chunk_size := get_long(fp);
chunk_arg  := get_word(fp);

if(chunk_size > ri^.buf_size) then ri^.buf_size := chunk_size;

if(feof(fp)) then break;

if(chunk_id = RoQ_INFO) then //	/* video info */
begin
ri^.width  := get_word(fp);
ri^.height := get_word(fp);
get_word(fp);
get_word(fp);
end
else
begin
if(chunk_id = RoQ_QUAD_VQ) then
begin
inc(ri^.num_frames);

SetLength(ri^.frame_offset,ri^.num_frames);
ri^.frame_offset[ri^.num_frames] := fpos;

end
else
if((chunk_id = RoQ_SOUND_MONO) or (chunk_id = RoQ_SOUND_STEREO)) then
begin
if(chunk_id = RoQ_SOUND_MONO) then ri^.audio_channels := 1 else ri^.audio_channels := 2;
ri^.num_audio_bytes := ri^.num_audio_bytes + chunk_size;
end;
fseek(fp, chunk_size, SEEK_CUR);
end;

end;

Result := 0;

end;

// Todo
{
static void apply_vector_2x2(roq_info *ri, int x, int y, roq_cell *cell)
static void apply_vector_4x4(roq_info *ri, int x, int y, roq_cell *cell)
static void apply_motion_4x4(roq_info *ri, int x, int y, unsigned char mv, char mean_x, char mean_y)
static void apply_motion_8x8(roq_info *ri, int x, int y, unsigned char mv, char mean_x, char mean_y)
}
//

Function roq_open(fname : pchar) : Proq_info;
var
F : PFile;
RI : Proq_info;
i : Integer;
begin

if not fileexists(fname) then
begin
messagebox(0,pchar('Error opening ' + fname),'Error Opening File',0);
Result := nil;
exit;
end;

AssignFile(F^,fname);
Reset(F^,1);
   {
try
ri := AllocMem(sizeof(troq_info));
except
messagebox(0,'Error alocating memory','Memory Error',0);
Result := nil;
exit;
end;    }
ri := nil;

ri^.FP := F;

if (roq_parse_file(f, ri) = 1) then
begin
Result := nil;
exit;
end;

ri^.stream_length := (ri^.num_frames * 1000) div 30;

for i := 0 to 127 do
begin
ri^.snd_sqr_arr[i]       := i * i;
ri^.snd_sqr_arr[i + 128] := -(i * i);
end;

for i := 0 to 1 do
begin
SetLength(ri^.y[i],ri^.width * ri^.height);
SetLength(ri^.u[i],ri^.width * ri^.height div 4);
SetLength(ri^.v[i],ri^.width * ri^.height div 4);
end;

ri^.buf_size := ri^.buf_size * 2;
SetLength(ri^.buf,ri^.buf_size);

ri^.audio_buf_size := 0;
ri^.audio          := nil;
ri^.frame_num      := 0;
ri^.aud_pos        := ri^.roq_start;
ri^.vid_pos        := ri^.roq_start;

Result := Ri;

end;

Procedure roq_close(RI : Proq_info);
var
I : integer;
begin

if ri = nil then exit;
CloseFile(ri^.fp^);

for i := 0 to 1 do
begin
Setlength(ri^.y[i],0);
Setlength(ri^.u[i],0);
Setlength(ri^.v[i],0);
end;
Setlength(ri^.buf,0);
FreeAndNil(Ri);
end;

//Todo
{
int roq_read_frame(roq_info *ri) // Needs todo procedures above made
}
//

Function roq_read_audio(Ri : Proq_info) : integer;
var
FP : PFile;
chunk_id,chunk_arg : integer;
chunk_size : longint;
i,snd_left,snd_right : integer;
begin
FP         := ri^.FP;
chunk_id   := 0;
chunk_arg  := 0;
chunk_size := 0;

Fseek(FP,ri^.aud_pos,SEEK_SET);
ri^.audio_size := 0;

while not feof(FP) do
begin
chunk_id   := get_word(fp);
chunk_size := get_long(fp);
chunk_arg  := get_word(fp);
if feof(fp) or (chunk_id = RoQ_SOUND_MONO) or (chunk_id = RoQ_SOUND_STEREO) then break;
fseek(fp, chunk_size, SEEK_CUR);
end;

if (ri^.audio_buf_size < chunk_size * 2) then
begin
setlength(ri^.audio,0);
ri^.audio_buf_size := chunk_size * 3;
setlength(ri^.audio,ri^.audio_buf_size);
end;

if(chunk_id = RoQ_SOUND_MONO) then
begin
ri^.audio_size := chunk_size * 2;
snd_left       := chunk_arg;
for i := 0 to chunk_size-1 do
begin
 snd_left             := snd_left + ri^.snd_sqr_arr[fgetc(fp)];
 ri^.audio[i * 2]     := snd_left and $ff;
 ri^.audio[i * 2 + 1] := (snd_left and $ff00) shr 8;
end;
ri^.aud_pos := ftell(fp);
Result := 1;
exit;
end;

if(chunk_id = RoQ_SOUND_STEREO) then
begin
ri^.audio_size := chunk_size * 2;
snd_left       := (chunk_arg and $FF00);
snd_right      := (chunk_arg and $FF) shl 8;
i := 0;
while i < chunk_size do
begin
 snd_left             := snd_left + ri^.snd_sqr_arr[fgetc(fp)];
 snd_right            := snd_right + ri^.snd_sqr_arr[fgetc(fp)];
 ri^.audio[i * 2]     := snd_left and $ff;
 ri^.audio[i * 2 + 1] := (snd_left and $ff00) shr 8;
 ri^.audio[i * 2 + 2] := snd_right and $ff;
 ri^.audio[i * 2 + 3] := (snd_right and $ff00) shr 8;
inc(i,2);
end;
ri^.aud_pos := ftell(fp);
Result := 1;
exit;
end;

ri^.aud_pos := ftell(fp);
Result := 0;
end;


end.
